/**************************************************************

cmdparser.c (C-Munipack project)
Command parser
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************

This code follows the implementation of command parser in axmail program:
 axMail - Copyright 1996 Heikki Hannikainen, OH7LZB <hessu@pspt.fi>
 
**************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "comfun.h"
#include "cmdparser.h"

/*
 *	Parse c-style escapes (neat to have!)
 */
 
static char *parse_string(char *str)
{
	char *cp = str;
	unsigned long num;

	while (*str != '\0' && *str != '\"') {
		if (*str == '\\') {
			str++;
			switch (*str++) {
			case 'n':
				*cp++ = '\n';
				break;
			case 't':
				*cp++ = '\t';
				break;
			case 'v':
				*cp++ = '\v';
				break;
			case 'b':
				*cp++ = '\b';
				break;
			case 'r':
				*cp++ = '\r';
				break;
			case 'f':
				*cp++ = '\f';
				break;
			case 'a':
				*cp++ = '\007';
				break;
			case '\\':
				*cp++ = '\\';
				break;
			case '\"':
				*cp++ = '\"';
				break;
			case 'x':
			    str--;
				num = strtoul(str, &str, 16);
				*cp++ = (char) num;
				break;
			case '0':
			case '1':
			case '2':
			case '3':
			case '4':
			case '5':
			case '6':
			case '7':
			    str--;
				num = strtoul(str, &str, 8);
				*cp++ = (char) num;
				break;
			case '\0':
				return NULL;
			default:
				*cp++ = *(str - 1);
				break;
			};
		} else {
			*cp++ = *str++;
		}
	}
	if (*str == '\"')
		str++;		/* skip final quote */
	*cp = '\0';		/* terminate string */
	return str;
}

// Parse 'cmd' string to the list of strings. Original string is divided into
// command and its arguments. First item of resulting 'argv' list points to the initial
// character of command, and next items points to the initial characters of arguments.
// Last item in the list is NULL. Only first 31 arguments are parsed.
int parse_args(char *argv[], char *cmd)
{
	int ct = 0;
	int quoted;

	while (ct < 31) {
		quoted = 0;
		while (*cmd && isspace(*cmd))
			cmd++;
		if (*cmd == 0)
			break;
		if (*cmd == '"') {
			quoted++;
			cmd++;
		}
		argv[ct++] = cmd;
		if (quoted) {
			if ((cmd = parse_string(cmd)) == NULL)
				return 0;
		} else {
			while (*cmd && !isspace(*cmd))
				cmd++;
		}
		if (*cmd)
			*cmd++ = 0;
	}
	argv[ct] = NULL;
	return ct;
}

// Prints specified prompt to stdout and then waits for user input on stdin.
// Returns number of bytes read (CR or LF characters are stripped out).
int getstr(const char *prompt, char *buf, int buflen)
{
	char *p;

	// Print prompt
	printf("%s", prompt);
	fflush(stdout);
	// Read user input
	fgets(buf, buflen, stdin);
	// Strip CR and LF
	p = strchr(buf, '\n');
	if (p != NULL) *p = 0; 
	p = strchr(buf, '\r');
	if (p != NULL) *p = 0;
	// Return number of characters
	return (int)strlen(buf);
}
