/**************************************************************

matread.c (C-Munipack project)
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#define _NO_OLDNAMES
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <expat.h>

#include "comfun.h"
#include "match.h"
#include "matfun.h"

#define BUFFSIZE 8192

static void XMLCALL start(void *data, const char *el, const char **attr)
{
  int i;
  double x,y;

  if ((strcmp(el,"s")==0)||(strcmp(el,"S")==0)) {
	  x = 0.0; y = 0.0;
      for (i = 0; attr[i]; i += 2) {
	      if ((strcmp(attr[i],"x")==0)||(strcmp(attr[i],"X")==0)) { x = atof(attr[i+1]); }
	      if ((strcmp(attr[i],"y")==0)||(strcmp(attr[i],"Y")==0)) { y = atof(attr[i+1]); }
      }
      mmatch_x1[mmatch_c1] = x;
      mmatch_y1[mmatch_c1] = y;
      mmatch_c1++;
  }
}

static int XMLCALL UnknownEncodingHandler(void *data,const XML_Char *encoding,XML_Encoding *info)
{
 int i;
 for (i = 0; i < 256; ++i) info->map[i] = i;
 info->data = NULL;
 info->convert = NULL;
 info->release = NULL;
 return XML_STATUS_OK;
}

int ReadRef(const char *filename)
{
  FILE *f;
  char *pt, ext[8];
  int i, w1, row;
  double u1, v1;
  char buf1[MAXLINE];
  int done, len;
  XML_Parser p;
  char Buff[BUFFSIZE];

  // Extract file's extension 
  pt = strrchr(filename,'.');
  if (pt) {
	for (i=1;(pt[i])&&(i<7);i++) { ext[i-1]=tolower(pt[i]); }
	ext[i-1] = 0;
  } else {
	  strcpy(ext, "");
  }

  f = fopen(filename,"r");
  if (!f) return MAT_CANNOT_OPEN_REF;

  if (!strcmp(ext,"xml")) {
	// Reads catalogue file 
    mmatch_c1 = 0;
  	p = XML_ParserCreate(NULL);
	XML_SetElementHandler(p, start, NULL);
  	XML_SetUnknownEncodingHandler(p, UnknownEncodingHandler, NULL);
  	do {
    	len = (int) fread(Buff, 1, BUFFSIZE, f);
    	if (ferror(f)) return -2;
    	done = feof(f);
    	if (XML_Parse(p, Buff, len, done) == XML_STATUS_ERROR) {
	    	// Parse error 
	    	sprintf(Buff,"Parse error at line %d:\n%s\n",
                XML_GetCurrentLineNumber(p),
              	XML_ErrorString(XML_GetErrorCode(p)));
        	PrintOut(0, Buff);
        	return MPK_INVALID_DATA;
    	}
  	} while (!done);
	fclose(f);
	return 0;
  }
  
  fgets(buf1,MAXLINE,f);
  if (strstr(buf1, "NL") && strstr(buf1, "NX") && strstr(buf1, "NY")) {
	/* Reads photometry file */
	fseek(f, 0, SEEK_SET);
	row = 0; mmatch_c1 = 0;
	while ((mmatch_c1<MAXSTARS)&&(fgets(buf1,MAXLINE,f)!=NULL)) {
		if ((row>=3)&&((row%3)==0)) {
			sscanf(buf1,"%d %lf %lf",&w1,&u1,&v1);
        	mmatch_x1[mmatch_c1]  = u1;
        	mmatch_y1[mmatch_c1]  = v1;
        	mmatch_c1++;
    	}
       	row++;
   	}
    fclose(f);
    return 0;
  }
  
  /* Normal return */
  fclose(f);
  return MPK_UNKNOWN_FORMAT;
}

int ReadSrc(const char *filename)
{
  FILE *f;
  size_t aux;
  int w2, row;
  double u2, v2;
  char buf1[MAXLINE];

  f = fopen(filename,"r");
  if (!f) return MPK_CANNOT_OPEN_SRC;

  mmatch_c2 = 0; row = 0;
  while ((mmatch_c2<MAXSTARS)&&(fgets(buf1,MAXLINE,f)!=NULL)) {
      if (row==0) {
          mmatch_line1 = malloc(strlen(buf1)+1);
          strcpy(mmatch_line1,buf1);
          // Umazat znaky CR+LF na konci radku
          aux = strlen(mmatch_line1);
          if ((aux>0)&&((mmatch_line1[aux-1]==13)||(mmatch_line1[aux-1]==10)))
            mmatch_line1[aux-1]=0;
          else if ((mmatch_line1[aux]==13)||(mmatch_line1[aux]==10))
            mmatch_line1[aux]=0;
      } else
      if (row==1) {
          mmatch_line2 = malloc(strlen(buf1)+1);
          strcpy(mmatch_line2,buf1);
          // Umazat znaky CR+LF na konci radku
          aux = strlen(mmatch_line2);
          if ((aux>0)&&((mmatch_line2[aux-1]==13)||(mmatch_line2[aux-1]==10)))
            mmatch_line2[aux-1]=0;
          else if ((mmatch_line2[aux]==13)||(mmatch_line2[aux]==10))
            mmatch_line2[aux]=0;
      } else
	  if ((row>=3)&&((row%3)==0)) {
          mmatch_l1[mmatch_c2] = malloc(strlen(buf1)+1);
	      strcpy(mmatch_l1[mmatch_c2],buf1);
		  sscanf(buf1,"%d %lf %lf",&w2,&u2,&v2);
      } else
      if ((row>=3)&&((row%3)==1)) {
	      mmatch_l2[mmatch_c2] = malloc(strlen(buf1)+1);
	      strcpy(mmatch_l2[mmatch_c2],buf1);
          mmatch_x2[mmatch_c2] = u2;
          mmatch_y2[mmatch_c2] = v2;
          mmatch_c2++;
      }
      row++;
  }
  fclose(f);  
  return 0;
}
