/**************************************************************

matsolve.c (C-Munipack project)
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/
#define _NO_OLDNAMES

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "comfun.h"
#include "quicksort.h"
#include "cholesky.h"
#include "robmean.h"
#include "match.h"
#include "matfun.h"
#include "matsimple.h"

static void Init1(int i1)
// Construct list of stars in image 1 and sorts them with respect of distance from the star i1.
{
    int i;
    double xc, yc, dx, dy;

    xc = mmatch_x1[i1];
    yc = mmatch_y1[i1];
    for (i=0; i<mmatch_n1; i++) {
	    mmatch_idr1[i] = i;
	    if (i==i1) {
	        mmatch_r1[i] = -1.0; 
    	} else {
	    	dx = mmatch_x1[i]-xc;
	    	dy = mmatch_y1[i]-yc;
	    	mmatch_r1[i] = dx*dx + dy*dy;
    	}
    }
    QuickSortD(mmatch_r1,mmatch_n1,(void **)mmatch_idr1);
}

static void Init2(int j1)
// Construct list of stars in image 2 and sorts them with respect of distance from the star j1.
{
    int j;
    double xc, yc, dx, dy;

    xc = mmatch_x2[j1];
    yc = mmatch_y2[j1];
    for (j=0; j<mmatch_n2; j++) {
	    mmatch_idr2[j] = j;
	    if (j==j1) {
	        mmatch_r2[j] = -1.0; 
    	} else {
	    	dx = mmatch_x2[j]-xc;
	    	dy = mmatch_y2[j]-yc;
	    	mmatch_r2[j] = dx*dx + dy*dy;
        }
    }
    QuickSortD(mmatch_r2,mmatch_n2,(void **)mmatch_idr2);
}

static int Resrov(int n, double c, double mirror, double axi, double ayi, double a2, double *b, double *m)
/* Resi soustavu rovnic a vypocte vyslednou transformacni matici M */
{
	double a[5*4];

	A(0,0) = n*c;
	A(1,0) = 0.0;
	A(2,0) = c*axi;
	A(3,0) = c*ayi;
	A(0,1) = 0.0;
	A(1,1) = A(0,0);
	A(2,1) = mirror*c*ayi;
	A(3,1) = -mirror*c*axi;
	A(0,2) = A(2,0);
	A(1,2) = A(2,1);
	A(2,2) = c*a2;
	A(3,2) = 0.0;
	A(0,3) = A(3,0);
	A(1,3) = A(3,1);
	A(2,3) = 0.0;
	A(3,3) = A(2,2);
	A(4,0) = c*b[0];
	A(4,1) = c*b[1];
	A(4,2) = c*b[2];
	A(4,3) = c*b[3];
	if (Cholesky(4,a)) 
	    return 1;
	if (fabs(A(4,2)*A(4,2)+A(4,3)*A(4,3)-1.0)>0.2) 
	    return 1;

	M(0,0) = A(4,2);
    M(0,1) = mirror*A(4,3);
    M(0,2) = A(4,0);
    M(1,0) = -mirror*A(4,3);
    M(1,1) = A(4,2);
    M(1,2) = A(4,1);
    M(2,0) = 0.0;
    M(2,1) = 0.0;
    M(2,2) = 1.0;
	return 0;
}

static int FindTrafo(int ns, double zoom, double *m, int *mstar, double *sumsq, double *tol)
/* Find coefficients of transformation */
{
      int    i,j,j0,nn,id1,id2;
      double s,axi,ayi,a2,xx1,yy1,xx2,yy2,xx,yy,tol2,dr,rmin,p[4];
      double dx,dy,mirror;

      // Set output to defaults
      mirror = 1;
	  memset(m, 0, 9*sizeof(double));
	  M(0,0) = M(1,1) = M(2,2) = 1.0;
	  *mstar = 0;
	  *sumsq = *tol = 0.0;
	  
      // We need at least one stars
      if (ns<1)
          return 0;		
          
      // Zero-iteration 
      axi = ayi = 0.0;
      for (i=0; i<ns; i++) {
	      id1  = mmatch_id1[i];
	      id2  = mmatch_id2[i];
	      dx   = mmatch_x2[id2] - mmatch_x1[id1];
	      dy   = mmatch_y2[id2] - mmatch_y1[id1];
	      if ((dx*dx+dy*dy)>mmatch_maxoffset*mmatch_maxoffset)
				return 0;
	      axi += dx;
	      ayi += dy;
      }
	  M(0,2) = axi/ns;
	  M(1,2) = ayi/ns;

	  // Only one star	  
	  if (ns==1) {
		  *mstar = 1;
		  *sumsq = 0;
		  *tol = 0.01;
		  return 1;
	  }
      
      // Residuals 
      s = 0.0;
      for (i=0;i<ns;i++) {
	      xx1 = mmatch_x1[mmatch_id1[i]];
	      yy1 = mmatch_y1[mmatch_id1[i]];
	      Trafo(m,xx1,yy1,&xx,&yy);
	      xx2 = mmatch_x2[mmatch_id2[i]];
	      yy2 = mmatch_y2[mmatch_id2[i]];
	      s  += (xx-xx2)*(xx-xx2)+(yy-yy2)*(yy-yy2);
      }
      tol2 = fmax(0.0001, 3.0*mmatch_clip*mmatch_clip*(s/ns)+MACHEPS);
	  if (tol2>mmatch_maxoffset*mmatch_maxoffset)
		  return 0;

      // Reziduals for all stars 
      nn = 0; j0 = -1;
      p[0] = p[1] = p[2] = p[3] = axi = ayi = a2 = s = 0.0;
      for (i=0;i<mmatch_n1;i++) {
	      Trafo(m,mmatch_x1[i],mmatch_y1[i],&xx,&yy);
          // Find star with the smallest residues 
          rmin = MACHMAX;
          for (j=0;j<mmatch_n2;j++) {
	          dr =(xx-mmatch_x2[j])*(xx-mmatch_x2[j])+(yy-mmatch_y2[j])*(yy-mmatch_y2[j]);
	          if (dr<rmin) { rmin = dr; j0 = j; }
          }
          j = j0;
          if (rmin<=tol2) {
	          // Star was found 
	          s   += rmin;
    	      xx1  = mmatch_x1[i];
	          yy1  = mmatch_y1[i];
	          xx2  = mmatch_x2[j];
	          yy2  = mmatch_y2[j];
	          axi += xx1;
	          ayi += yy1;
	          a2  += xx1*xx1+yy1*yy1;
	          p[0]+= xx2;
	          p[1]+= yy2;
	          p[2]+= xx1*xx2+mirror*yy1*yy2;
	          p[3]+= yy1*xx2-mirror*xx1*yy2;
	          nn++;
          }
      }
      
      // We need at least one star
      if (nn<1)
	      return 0;
	      
	  // Find improved transformation matrix
      if (Resrov(nn,zoom,mirror,axi,ayi,a2,p,m)!=0)
	      return 0;
	      
	  // Compute improved tolerance
      tol2 = fmax(0.0001, 3.0*mmatch_clip*mmatch_clip*(s/nn)+MACHEPS);
	  if (tol2>mmatch_maxoffset*mmatch_maxoffset)
		  return 0;

      // Test the transformation, comute final tolerance and number of identified stars
      nn=0; s=0.0;
      for (i=0;i<mmatch_n1;i++) {
	      Trafo(m,mmatch_x1[i],mmatch_y1[i],&xx,&yy);
          // Find star with the smallest residues
          rmin = MACHMAX;
          for (j=0;j<mmatch_n2;j++) {
	          dr =(xx-mmatch_x2[j])*(xx-mmatch_x2[j])+(yy-mmatch_y2[j])*(yy-mmatch_y2[j]);
	          if (dr<rmin) { rmin = dr; }
          }
          if (rmin<tol2) {
	          s += rmin;
	          nn++;
          }
	}
	
	// Transformation is OK, if there are at least three stars identified
	if (nn<1) 
		return 0;
      
	// Normal return, transformation is OK
	*tol = fmax(0.0001, 3.0*mmatch_clip*mmatch_clip*(s/nn)+MACHEPS);
	*mstar = nn;          
	*sumsq = s;
	return 1;
}

static int TestMatch(int n, double *zoom)
/* Mame uz pozadovany pocet hvezd v polygonech id1 a id2 -> testuj polygon */
{
	int    i1, i2, i3, i, j, k, l, rep;
	double dx1, dx2, dy1, dy2, s, maxdev, sig;
	double a1[2], a2[2], b1[2], b2[2];
	double  xmed, xsig, xdev, xarc, prum;

	*zoom = 1.0;
	
	// We need at least one star
	if (n<1)
		return 0;
	
	if (n>=3) {
		// Calculation of ratio of similarity (scale)
		// We need at least two stars
		l = 0; rep = 0;
		for (i=0;i<n;i++) {
			for (j=i+1;j<n;j++) {
				i1  = mmatch_id1[i];
				i2  = mmatch_id1[j];
				dx1 = mmatch_x1[i2]-mmatch_x1[i1];
				dy1 = mmatch_y1[i2]-mmatch_y1[i1];
	            i1  = mmatch_id2[i];
	            i2  = mmatch_id2[j];
	            dx2 = mmatch_x2[i2]-mmatch_x2[i1];
	            dy2 = mmatch_y2[i2]-mmatch_y2[i1];
	            mmatch_k2[l] = sqrt((dx2*dx2+dy2*dy2)/(dx1*dx1+dy1*dy1));
	            if ((l==0)&&(mmatch_k2[l]<0.9)) rep = 1;
	            if (rep) mmatch_k2[l] = 1.0/mmatch_k2[l];
				l++;
			}
		}
		// Calculation of mean and median
		cmpack_robustmeanD(l,mmatch_k2,&xmed,&xsig);
    	// Calculation of absolute deviations 
	    for (i=0;i<l;i++) {
		    mmatch_dev[i] = fabs(mmatch_k2[i]-xmed);
		    if (mmatch_dev[i] > 0.2*xmed) 
		    	return 0;   			// deviation is too great (because it isn't from the true series)
	    }
		// Calculation of the median deviation 
		cmpack_robustmeanD(l,mmatch_dev,&xdev,&xsig);
		if (xdev > 0.05*xmed) 
			return 0;       			// Mean deviation is too great
		// Calculation of the aritmetical mean 
	    s = 0.0;
	    for (i=0;i<l;i++) 
	    	s += mmatch_k2[i];
	    prum = s/l;
    	// Find the greatest deviation into mean
	    j = 0; maxdev = -1.0;
	    for (i=0;i<l;i++) {
		    mmatch_dev[i] = (mmatch_k2[i]-prum)*(mmatch_k2[i]-prum);
	        if (mmatch_dev[i]>maxdev) {
		        maxdev = mmatch_dev[i];
		        j = i;
	        }
	    }
    	// Calculation of the new mean 
	    s = 0.0;
	    for (i=0;i<l;i++) 
	    	if (i!=j) s += mmatch_k2[i];
	    prum = s/(l-1);
    	// Estimation of the standard deviation 
	    s = 0.0;
	    for (i=0;i<l;i++) {
		    mmatch_dev[i] = (mmatch_k2[i]-prum);
	        if (i!=j) s += mmatch_dev[i]*mmatch_dev[i];
	    }
	    sig = sqrt(s/(l-2))+MACHEPS;
	    if (sig > 0.05*prum) 
	    	return 0;   				// Standard deviation is too great
	    if (fabs(prum-xmed)>=sig) 
	    	return 0; 					// Test for equality of median and mean */
	    // Compute zoom estimation
    	if (rep) *zoom = 1.0/xmed;
        	else *zoom = xmed;
    	// Test on the same arcs of N-polygon 
    	// We need at least three stars
	    l = 0;
		for (i=0;i<n;i++) {
			for (j=i+1;j<n;j++) {
			    for (k=j+1;k<n;k++) {
				    i1 = mmatch_id1[i];
				    i2 = mmatch_id1[j];
				    i3 = mmatch_id1[k];
				    a1[0] = mmatch_x1[i2]-mmatch_x1[i1];
				    a1[1] = mmatch_y1[i2]-mmatch_y1[i1];
				    b1[0] = mmatch_x1[i3]-mmatch_x1[i1];
				    b1[1] = mmatch_y1[i3]-mmatch_y1[i1];
				    i1 = mmatch_id2[i];
				    i2 = mmatch_id2[j];
				    i3 = mmatch_id2[k];
				    a2[0] = mmatch_x2[i2]-mmatch_x2[i1];
				    a2[1] = mmatch_y2[i2]-mmatch_y2[i1];
				    b2[0] = mmatch_x2[i3]-mmatch_x2[i1];
				    b2[1] = mmatch_y2[i3]-mmatch_y2[i1];
				    mmatch_k2[l] = Uhel(a1,b1)/Uhel(a2,b2);
				    if ((mmatch_k2[l]>2.0)||(mmatch_k2[l]<0.5)) 
				    	return 0; // devation onto 1.0 is too great 
				    l++;
			    }
		    }
	    }
	    // Calculation of mean and median of arcs 
		cmpack_robustmeanD(l,mmatch_k2,&xarc,&xsig);
    	// Calculation of the deviation into median 
	    for (i=0;i<l;i++) {
		    mmatch_dev[i] = fabs(mmatch_k2[i]-xarc);
	        if (mmatch_dev[i] > 0.2*xarc) 
	        	return 0;
	    }
		cmpack_robustmeanD(l,mmatch_dev,&xdev,&xsig);
	    if (xdev+MACHEPS > 0.2*xarc) 
	    	return 0;           // Mean deviation is too great 
	    if (fabs(xarc-1.0) > 3.0*(xdev+MACHEPS)) 
	    	return 0; 			// Median must be equal 1.0 
    }
      
	// All test are OK
    return 1;
}

int Simple(int *m_stars)
{
	int 	mstar, nstar, i, j, j0, ni, i1, i3, j1, j3, i4, j4;
	double 	dx, dy, xx, yy, dr, rr, rl, rg, rmax, dif, u1, u2, v1, v2, s;
	double	sumsq, tol2, zoom, m[9];
	char	msg[256];

	sprintf(msg,"Matching algorithm               : Sparse fields");
	PrintOut(0, msg);
	
	// Clear output
	*m_stars = 0;
    mmatch_offx = 0.0;
    mmatch_offy = 0.0;
    for (i=0; i<mmatch_c2; i++) 
    	mmatch_id[i] = -1;
	
	// Number of stars used for matching
	mmatch_n1 = (mmatch_c1>mmatch_maxstar ? mmatch_maxstar : mmatch_c1);
	mmatch_n2 = (mmatch_c2>mmatch_maxstar ? mmatch_maxstar : mmatch_c2);

	// For all stars on reference frame -> i1
	for (i1=0;i1<mmatch_n1;i1++) {
		Init1(i1);
	    mmatch_id1[0] = mmatch_idr1[0];
		// For all stars on input frame -> j1
		for (j1=0;j1<mmatch_n2;j1++) {
			dx = (mmatch_x2[j1] - mmatch_x1[i1]);
			dy = (mmatch_y2[j1] - mmatch_y1[i1]);
			
			// Check max. offset
			if ((dx*dx + dy*dy)>mmatch_maxoffset*mmatch_maxoffset) 
				continue;
			
			Init2(j1);
		    mmatch_id2[0] = mmatch_idr2[0];
		    ni = 1;
		    // Try matching using only one star
            if (TestMatch(1,&zoom)) {
				// Find transformation matrix 
			    if (FindTrafo(1,zoom,m,&mstar,&s,&tol2)) {
				    // Add transformation to the stack
			        StAdd(1,s,tol2,mstar,m);
				}
			}
			for (i3=1; i3<mmatch_n1; i3++) {
                // Find the star on input frame with the distance most close to i3 -> j3
                rr   = mmatch_r1[i3];
            	rg   = 1.3*rr;
                rmax = MACHMAX;
                j3  = -1;
                for (j=2; j<mmatch_n2 && mmatch_r2[j]<rg; j++) {
                    if (mmatch_idr2[j]!=mmatch_id2[0]) {
                    	dif = fabs(mmatch_r2[j]-rr);
                		if (dif<rmax) { 
                    		rmax = dif; 
                    		j3 = j; 
                    	}
                	}
                }
                if (j3<0) 
                	continue; // Next i3

                // Add i3 and j3 to the polygon 
                mmatch_id1[1] = mmatch_idr1[i3];
                mmatch_id2[1] = mmatch_idr2[j3];
                if (TestMatch(2,&zoom)) {
					// Find transformation matrix 
				    if (FindTrafo(2,zoom,m,&mstar,&s,&tol2)) {
					    // Add transformation to the stack
				        StAdd(2,s,tol2,mstar,m);
					}
				}
                
				ni = 2;
				for (i4=i3+1; i4<mmatch_n1 && ni<mmatch_maxstar; i4++) {
	                // Find the star on input frame with the distance most close to i3 -> j3
	                rr   = mmatch_r1[i4];
	                rl   = rr/1.3;
	            	rg   = rr*1.3;
	                rmax = MACHMAX;
	                j4  = -1;
	                for (j=2; j<mmatch_n2 && mmatch_r2[j]<rg; j++) {
	                    if (mmatch_r2[j]>rl && mmatch_idr2[j]!=mmatch_id2[0] && mmatch_idr2[j]!=mmatch_id2[1]) {
	                    	dif = fabs(mmatch_r2[j]-rr);
	                		if (dif<rmax) { 
	                    		rmax = dif; 
	                    		j4 = j; 
	                    	}
	                	}
	                }
	                if (j4<0) 
	                	continue; // Next i3
	
	                // Add i3 and j3 to the polygon 
	                mmatch_id1[ni] = mmatch_idr1[i4];
	                mmatch_id2[ni] = mmatch_idr2[j4];
				
	                // Calculation of the triangle coordinates (u1,v1) in the reference file
	                UV(mmatch_id1[ni-2],mmatch_id1[ni-1],mmatch_id1[ni],mmatch_x1,mmatch_y1,&u1,&v1);
	                // Calculation of the triangle coordinates (u2,v2) in the input file 
	                UV(mmatch_id2[ni-2],mmatch_id2[ni-1],mmatch_id2[ni],mmatch_x2,mmatch_y2,&u2,&v2);
	                if (((u1-u2)*(u1-u2)+(v1-v2)*(v1-v2))<0.0005) {
	                    // Transformation is found
	                    ni++;
	                    if (TestMatch(ni,&zoom)) {
							// Find transformation matrix 
						    if (FindTrafo(ni,zoom,m,&mstar,&s,&tol2)) {
							    // Add transformation to the stack
						        StAdd(ni,s,tol2,mstar,m);
							}
						}
	                    if (ni>=mmatch_nstar) {
	                        break;	// Next j2
	                    }
				    }
			    }
            }
        }
    }

	// Select from stack the best case
    StSelect(&mstar,&sumsq,&tol2,m);
    if (mstar==0) {
	    // No coincidences found
	    PrintOut(0, "Coincidences not found!");
	    return MAT_MATCH_NOT_FOUND;
    }
    
    // Matching OK
	sprintf(msg,"No. of stars used for matching   : %d", mstar);
	PrintOut(0, msg);
	sprintf(msg,"Sum of square in the best case   : %.4f", sumsq);
	PrintOut(0, msg);
	sprintf(msg,"Tolerance                        : %.2f", sqrt(tol2));
	PrintOut(0, msg);
	sprintf(msg,"Transformation matrix            : ");
	PrintOut(0, msg);
	sprintf(msg,"   %15.3f %15.3f %15.3f",M(0,0),M(0,1),M(0,2));
	PrintOut(0, msg);
	sprintf(msg,"   %15.3f %15.3f %15.3f",M(1,0),M(1,1),M(1,2));
	PrintOut(0, msg);
	sprintf(msg,"   %15.3f %15.3f %15.3f",M(2,0),M(2,1),M(2,2));
	PrintOut(0, msg);
	
	// Set cross-references for all stars on input frame
	nstar = 0;
    for (i=0; i<mmatch_c1; i++) {
	    // Compute estimated destination coordinates
        Trafo(m,mmatch_x1[i],mmatch_y1[i],&xx,&yy);
        // Select the nearest star 
        rr = MACHMAX; j0 = -1;
        for (j=0; j<mmatch_c2; j++) {
	        if (mmatch_id[j]<0) {
		        dx = xx-mmatch_x2[j];
		        dy = yy-mmatch_y2[j];
		        dr = (dx*dx+dy*dy);
		        if (dr<rr) { rr = dr; j0 = j; }
	        }
        }
        if ((j0>=0)&&(rr<tol2)) {
	        mmatch_id[j0] = i;
	        nstar++;
        }
    }
    *m_stars = nstar;
    mmatch_offx = M(0,2);
    mmatch_offy = M(1,2);

	return 0;
}
