/**************************************************************

attach.c (C-Munipack project)
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <math.h>
#include <fitsio.h>

#include "comfun.h"
#include "fitsfun.h"
#include "phot.h"
#include "phfun.h"
#include "attach.h"

int Attach(const char *nafile)
/* Reads ph_image data from fits file into memory buffer */
{
    int fstat, anyf, bitpix;
    fitsfile *fits;

    /* Open source file */
    fstat = 0;
    ffdkopn(&fits,nafile,READONLY,&fstat);
    if (fstat!=0) return fstat;

    /* Read frame attributes */
    fstat = fits_getjd(fits, &ph_image.jd);
    if (fstat!=0) {
 		ffclos(fits,&fstat);
		return fstat;
    }
	fits_gfilter(fits, ph_image.filter);
	fits_gobject(fits, ph_image.object);
	fits_gexp(fits, &ph_image.exptime);
	fits_gtemp(fits, &ph_image.temp);
	fits_gpar(fits, &ph_image.ncol, &ph_image.nrow, &bitpix);
    ph_image.filter[16] = 0; // Crop the filter name to 16 characters
    if (ph_fravgauto) {
	    ph_fravg = 1;
	    fits_gfravg(fits, &ph_fravg);
    }

    /* Check attributes */    
    if (ph_fravg<=0) {
	    PrintOut(0, "Number of frames averaged is <=0.");
	    return MPK_INVALID_PAR;
    }
    if ((ph_image.ncol<=0)||(ph_image.nrow<=0)) {
 		ffclos(fits,&fstat);
		return MPK_INVALID_SIZE;
    }
    if (ph_image.jd<=0) {
 		ffclos(fits,&fstat);
		return MPK_INVALID_DATE;
    }
    ph_image.jd += 0.5*ph_image.exptime/86400.0;

    /* Read image */
    if (ph_image.ccd)
        ph_image.ccd = (double*) realloc(ph_image.ccd,ph_image.nrow*ph_image.ncol*sizeof(double));
    else
        ph_image.ccd = (double*) malloc(ph_image.nrow*ph_image.ncol*sizeof(double));
    if (!ph_image.ccd) {
	    ffclos(fits,&fstat);
	    return MPK_MEMORY;
    }
    ffg2dd(fits,1,0,ph_image.ncol,ph_image.ncol,ph_image.nrow,ph_image.ccd,&anyf,&fstat);
    if (fstat!=0) {
	    ffclos(fits,&fstat);
	    return fstat;
    }

    /* Close ph_image file */
    ffclos(fits,&fstat);

    /* Normal return */
    return 0;
}
